var logger = require('logger').getLogger('noderpc'),
        util = require('util');

function HttpBinding(server) {
    var rpcMarshal = require('noderpc/marshal'),
        connect = require('connect'),
        api = {},
        boundServices = {};

    this.unbindService = function(serviceName) {
        if (boundServices[serviceName] === undefined) {
            throw new Error("Service '" + serviceName + "' not bound");
        }
        delete boundServices[serviceName];
        delete api[serviceName];
    }

    this.bindService = function(bindObject) {
        var serviceName = bindObject.serviceDefinition.name;
        if (boundServices[serviceName] !== undefined) {
            throw new Error("Service '" + serviceName + "' already bound");
        }
        boundServices[serviceName] = bindObject;
        api[serviceName] = bindObject.serviceDefinition.methods;
    }

    this.serveStatic = function(path) {
        server.use(
            connect.static(path || __dirname + '/../../../public')
        );
    }

    this.setLogger = function(log4js, level, format) {
        var logStream = {
            writable: true,
            write: function(string) {
                log4js[level](string.substring(0, string.length - 1));
            },
            end: function() {},
            destroy: function() {}
        }

        server.use(connect.logger({
            format: format,
            stream: logStream
        }));
    }

    function processCall(service, methodName, req, res) {
        try {
            var paramDef = service.serviceDefinition.methods[methodName].params,
                formEncoded = (req.headers['content-type'] != 'application/json'),
                // Marshall request arguements
                args = rpcMarshal.fromHttpParams(req.body, paramDef, formEncoded);

            // Append success callback
            args.push(function(result) {
                res.writeHead(200, {'Content-Type': 'application/json'});
                if (result !== undefined) {
                    res.write(JSON.stringify(result));
                }
                res.end();
            });
            // Append error callback
            args.push(function(error) {
                if (typeof error != 'string') {
                    error = error.message;
                }
                res.writeHead(500, {'Content-Type': 'text/html'});
                res.write(error);
                res.end();
            });

            // Call native object method
            // bindObject = thisContext
            // bindObject[name] = native method
            service[methodName].apply(service, args);
        }
        catch (error) {
            if (typeof error != 'string') {
                error = error.message;
            }
            res.writeHead(500, {'Content-Type': 'text/html'});
            res.write(error);
            res.end();
        }
    }

    function requestHandler(req, res, next) {
        var levels,
            serviceName, methodName,
            service;
        levels = req.url.split('/');
        levels.shift();
        serviceName = levels.shift(),
        methodName = levels.shift();

        if (levels.length > 0 ||
            methodName === undefined ||
            boundServices[serviceName] === undefined) {
            next();
        }
        service = boundServices[serviceName];
        if (typeof service[methodName] !== 'function' ||
            boundServices[serviceName] === undefined) {
            next();
        }
        processCall(service, methodName, req, res);
    }

    this.start = function(port, hostname) {
        server.use(
            connect.router(function(app) {
            app.get('/api', function(req, res, next) {
                res.writeHead(200, {'Content-Type': 'text/html'});
                res.write(JSON.stringify(api));
                res.end();
            });
            app.get('/heartbeat', function(req, res, next) {
                res.writeHead(200, {'Content-Type': 'text/html'});
                res.write("OK");
                // Don't end request
            });
        }));
        server.use(requestHandler);
        server.listen(port, hostname);
    }

    this.stop = function() {
        server.close();
    }
}

module.exports = function(connectServer) {
    connectServer.use(require('connect').bodyParser());
    return new HttpBinding(connectServer);
}